
#include <Arduino.h>
#include <avr/pgmspace.h> 
//#include "TimerOne_x.h"
#include "kasoku.h"
#include "VVVF_sound.h"
#include "motor_func.h"

#include "shu1_8k_IMA.h"
#include "shu2_8k_IMA.h"
#include "shu3_8k_IMA.h"
#include "shu4_8k_IMA.h"
#include "whistle_short2_IMA.h"
#include "whistle_middle_IMA.h"
#include "whistle_long2_IMA.h"


// REF_WAVE_FREQ = 1000000 / (WAV_LENGTH * 62.5)
//#define REF_WAVE_FREQ 12.618
//#define REF_WAVE_FREQ 15.625
//#define SHIFT_PWM 5
#define SHIFT_PWM 5

//#define  COUNT_CONSTANTSPEED  165 /* 33d=1sec  */
#define  COUNT_CONSTANTSPEED   33 /* 33d=1sec  */
#define  THRESHOLD_RATIO       40 /* ノッチオフの加減速しきい値 */

//VVVF
int uplimit; //VVVF配列の上限
int kasoku[15][8];//VVVF配列
//#define CARRER_FREQ 22000
#define CARRER_FREQ 16000


int current_period1 =(double)1000000 / CARRER_FREQ;//VVVF周期

volatile unsigned int pwm_state1 = 0;
volatile unsigned int function_state1 = 0;
volatile int function_num = 0;
unsigned int interval_count = 3000;

unsigned int pwm_add1 = 0;

unsigned int pwm_shift = 8;
//unsigned int pwm_shift = 0;

//IMA ADPCM関連
IMADEC imad;
uint16_t nBlockAlign = 252;

int current_duty_a = 0;


int gAccRatio = 20;
int gDecRatio = 20;
int gNotchSpdCount = 0;

uint8_t SoundSwitch = 0;
uint8_t SoundState = 1;
uint8_t SoundNotch = 0;
uint8_t SoundVolume = 0;
long gVVVFLPF_buf = 0;

float PrevSpeed = 0;

boolean accel_flag = false;
boolean draft_flag = false;

//割り込みのひな形
void vvvf_int1();

void VVVF_Setup()
{
  //debug
  //pinMode(7, OUTPUT);
  
  //PWM出力ピン D9を出力にセット
  pinMode(VVVF_SOUND_PIN, OUTPUT);
  //Timer1使用
  //Timer1.initialize();
  //初期設定
  //Timer1.pwm(VVVF_SOUND_PIN,0,0);
  //Timer1.attachInterrupt(vvvf_int1,current_period1 / 2);//interruptの設定    

  //IMA ASPCM 関連
  uint8_t gSound_played = 0; 
  nBlockAlign = (uint16_t)(readSoundRom(gSound_played, 33) << 8 | readSoundRom(gSound_played, 32)) - 4;//周波数によってブロックサイズが違う。

  //31KHzに設定（初期設定でPhaseCorrectPWMモードになっていると思う）
  //D9,D10  キャリア周期:31kHz(分周無し）
  TCCR1B &= B11111000;
  TCCR1B |= B00000001;

  //出力の設定
  if(VVVF_SOUND_PIN == 9)
  {
    TCCR1A |= _BV(COM1A1);
  }
  else //10
  {
    TCCR1A |= _BV(COM1B1);
  }
  
  //interrupt設定
  TIMSK1 = _BV(TOIE1); 
}

void VVVF_SetCV(uint8_t iNo, uint8_t inData)
{
  switch(iNo)
  {
	case 3:
		
		if( inData > 16)
		{
			gAccRatio = 255;
		}
		else
		{
			gAccRatio = inData * 16;
		}
	
		if( gAccRatio <= 0)
	  	{
	  		gAccRatio = 1;
	  	}
		
		break;
	case 4:
		if( inData > 16)
		{
			gDecRatio = 255;
		}
		else
		{
			gDecRatio = inData * 16;
		}
		
		if( gDecRatio <= 0)
	  	{
	  		gDecRatio = 1;
	  	}
  		break;
  case 47:
    if(inData > 9)
    {
      inData = 9;
    }
    SoundSwitch = inData;
    break;
  case 48:
    if(inData > 1)
    {
      inData = 1;
    }
    SoundState = inData;
    break;
  case 49:
    if(inData > 2)
    {
      inData = 2;
    }
    SoundNotch = inData;
    break;
  case 50:
    if( inData > 8)
    {
      inData = 8;//0:min(no sound) ~ 8:max
    }
    SoundVolume = inData;
    break;
  }
}

void VVVF_Cont(int inPWMFreq, uint8_t inF2Flag)
{
	uint16_t aPWMRef = 0;
  static uint8_t inF2Flag_static = 0x00; 

	//LPF
	aPWMRef = MOTOR_LPF(inPWMFreq, 32, &gVVVFLPF_buf);
	
	//int to float
	float instruct_speed = (float)(aPWMRef * 4);//change from 256 to 1023
	float speed_step = 0;
	
  if(inPWMFreq == 0)
  {
    accel_flag = false;
    PrevSpeed = 0;
  }
  else if( (PrevSpeed - instruct_speed) < -1 )
  {
    //Acceleration
    accel_flag = true;
    speed_step = (abs(instruct_speed - PrevSpeed) / gAccRatio) + 1;
  	
    PrevSpeed += speed_step ;
    if (PrevSpeed > instruct_speed)
    {
      PrevSpeed = instruct_speed;
    }
  	
  	//ノッチオフの音を止めるタイマ
  	gNotchSpdCount = COUNT_CONSTANTSPEED;
  }
  else if( (PrevSpeed - instruct_speed) > 1 )
  {
    //Deceleration
    accel_flag = false;
    speed_step = (abs(instruct_speed - PrevSpeed) / gDecRatio) + 1;
    PrevSpeed -= speed_step;
    if (PrevSpeed < instruct_speed)
    {
      PrevSpeed = instruct_speed;
    }
  	
  	gNotchSpdCount = 0;
  }
  else
  {
  	//Constant speed
  	
  	
  	if( gNotchSpdCount > 0)
  	{
  		gNotchSpdCount--;
	    accel_flag = true;
  	}
  	else
  	{
	    accel_flag = false;
  	}
  	
  }

  /* 
  ノッチオフ無効(0)時はアクセルフラグを常時有効（減速でオフにさせない) 
  ノッチオフ自動(1)時はアクセルフラグを自動制御(加速終了後指定時間でオフ)
  ノッチオフ手動(2)時はアクセルフラグを手動操作
  */
  if (SoundNotch == 2 )
  {
    if ((inF2Flag & 0x01) == 0)
    {
      accel_flag = true;
    }
    else
    {
      accel_flag = false;
    }
    
  }
  else
  {
    //何もしない
  }


  //ファンクション
  //変更されていたら鳴らす
  uint8_t temp_flag = inF2Flag_static ^ inF2Flag;
  switch(temp_flag)
  {
    case 0x01:
      //F03 ノッチなので無視
      break;   
    case 0x02:
      //F04
      function_num = 4;
      function_state1 = 0;
      break;   
    case 0x04:
      //F05
      function_num = 5;
      function_state1 = 0;
      break;   
    case 0x08:
      //F06
      function_num = 6;
      function_state1 = 0;
      break;   
    default:
      break;   
  }
  //最後に値をStaticに入れておく
  inF2Flag_static = inF2Flag;

  VVVF_Freq((int)PrevSpeed);

	/* 加減速管理 */
/*
  if((PrevSpeed != 0) && ((accel_flag == true) || ( SoundNotch == 0)))
  {
    if(SoundVolume > (8 - pwm_shift))
    {
      pwm_shift --;
    }
  }
  else
  {
    if(pwm_shift < 8)
    {
      pwm_shift ++;
    }
  }
*/

  //音量関連
  //最初の過渡対応
  if(SoundVolume > (8 - pwm_shift))
  {
    pwm_shift --;
  }
  //シュー音のOnOffはdraft_flagを渡して管理
  if((PrevSpeed != 0) && ((accel_flag == true) || ( SoundNotch == 0)))
  {
    draft_flag = true;
  }
  else
  {
    draft_flag = false;
  }
}

int VVVF_Freq( int inPWMFreq)
{
  //VVVF Freq部分
  double pwm_freq1 = 440;
  double pwm_freq2 = 440;
  
  for(int i = 0;i < uplimit;i++)
  {
    if ( (kasoku[i][0] <= inPWMFreq ) and ( inPWMFreq <= kasoku [i][1]))
    {
      //音色側
      //周波数１
      pwm_freq1 = (double)kasoku[i][2]
               + (double)(kasoku[i][3] - kasoku[i][2]) / (double)(kasoku [i][1] - kasoku [i][0])
               * (double)(inPWMFreq - kasoku [i][0]);
    }  
  }

 
  //pwm_add1 = (pwm_freq1 * pow(2,SHIFT_PWM)) / REF_WAVE_FREQ1;
  interval_count =  16000.0 * 20.0  / pwm_freq1;
  
  return 0;
}

void VVVF_Init()//160917 fujigaya2 
{
  int cv47_Value = SoundSwitch;
  int cv48_Value = SoundState;
  //VVVF配列読み込み
  int kasoku_offset =  (int)pgm_read_word_near(&progmem_ref_kasoku[cv47_Value]);
  uplimit = (int)pgm_read_word_near(&progmem_ref_kasoku[cv47_Value+1]) - kasoku_offset;
  for(int i = 0;i < uplimit;i++)
  {
    for (int j = 0;j < 8;j++)
    {
      kasoku[i][j] =  (int)pgm_read_word_near(&progmem_kasoku [i + kasoku_offset][j]); 
    }
  }
}

//void vvvf_int1()
ISR(TIMER1_OVF_vect)
{
  //波形生成用interrupt
  //static boolean flg = true;
  static unsigned char flg = 0;  
  static unsigned char sound_num = 0;
  static unsigned int prev_value = 128;

  if(flg == 0x03)
  {
    if (VVVF_SOUND_PIN == 9)
    {
      OCR1A=prev_value>>pwm_shift;
    }
    else//(VVVF_SOUND_PIN == 10)
    {
      OCR1B=prev_value>>pwm_shift;
    }
    flg = 0x00;
    return;
  }
  else if(flg == 0x01)
  {
    flg++;
  }
  else
  {
    flg++;
    return;
  }
  
  unsigned int temp_value = 128;
 
  if((pwm_state1 > interval_count))
  {
    //インターバル時間を超えたら０に戻す
    pwm_state1 = 0;
  }

  //シュー音の管理
  if(pwm_state1 == 0)
  {
    //0の時に音色を進める
    sound_num ++; 
  }
  
  if(function_num == 0)
  {
    //シュー音　4気筒　
    temp_value = readWavData(sound_num & 0x03,pwm_state1);
  }
  else
  {
    //Function 4-6
    temp_value = readWavData(function_num,function_state1);
    function_state1 = 1;
  }
  
  //線形補間
  unsigned int temp_value2 = (temp_value + prev_value) >> 1;
  prev_value = temp_value;
  if (VVVF_SOUND_PIN == 9)
  {
    OCR1A=temp_value2>>pwm_shift;
  }
  else//(VVVF_SOUND_PIN == 10)
  {
    OCR1B=temp_value2>>pwm_shift;
  }

  //シュー音のOnOff
  if(draft_flag == true)
  {
    pwm_state1 ++;
  }
}

//元々のWAVで"pgm_read_byte_near(&WAV_DATA1 [temp_state])";に当たるところ
byte readWavData(int gSound_played,unsigned int pwm_state1)
{
    static uint8_t hilow = LOW;   // add aya:最初は下位
    static uint16_t ndata = 0;
    static uint16_t sample = 64;
    static uint16_t gSound_Length =0;
    static uint8_t prev_value = 128;  
   
    //制御関連
    if (pwm_state1 == 0)
    {
      // 初期設定
      hilow = LOW;
      ndata = 0;
      sample = 64;
      imad.val = (int16_t)(readSoundRom(gSound_played, 61) << 8 | readSoundRom(gSound_played, 60));   // Get 1st sample value of the block
      imad.idx = readSoundRom(gSound_played, 62);
      gSound_Length = (uint16_t)(readSoundRom(gSound_played, 57) << 8 | readSoundRom(gSound_played, 56)) - 13 * 20; // RIFF No56,No57
      //nBlockAlign = (uint16_t)(readSoundRom(gSound_played, 33) << 8 | readSoundRom(gSound_played, 32)) - 4;//周波数によってブロックサイズが違う。      
      return 128;
    }
    else if  (sample >= gSound_Length - 50)     //差分は要検討
    {
      //音再生終了のため何もしない
      function_num = 0;//音終了のサイン
      return 128;
    }
    else if (ndata == nBlockAlign)    // 11/9 1 block 再生終わったら、val,idxを再セットする //16kHz?
    {
      ndata = 0;
      imad.val = (int16_t)(readSoundRom(gSound_played, sample + 1) << 8 | readSoundRom(gSound_played, sample)); // Get 1st sample value of the block
      imad.idx = readSoundRom(gSound_played, sample + 2);
      sample += 4;
      return prev_value;
    }
    
    //音再生
    uint8_t value = ima_decode(&imad, readSoundRom(gSound_played, sample), hilow);  // Sound extension by MECY 2017/08/02
    prev_value = value;    
    if (hilow == HIGH)   // add aya : IMA ADPCM は1byteに4bit*2のデータがあるからね。
    {
      hilow = LOW;
      sample++;     // 再生アドレスをインクリメント
      ndata++;      // IMA ADPMC 1BLOCKカウントをインクリメント
    }
    else
    {
      hilow = HIGH;
    }
    return value;
}

//------------------------------------------------------------------------------
//  ROMに配置されている変数から値を取り出す処理
//------------------------------------------------------------------------------
byte readSoundRom(uint8_t inNo, uint16_t inOffset)
{
  switch (inNo) {
    case 0:
      return pgm_read_byte(&shu1_8k_IMA_data[inOffset]);
      break;
    case 1:
      return pgm_read_byte(&shu2_8k_IMA_data[inOffset]);
      break;
    case 2:
      return pgm_read_byte(&shu3_8k_IMA_data[inOffset]);
      break;
    case 3:
      return pgm_read_byte(&shu4_8k_IMA_data[inOffset]);
      break;
    case 4:
      return pgm_read_byte(&whistle_short2_IMA_data[inOffset]);
      break;
    case 5:
      return pgm_read_byte(&whistle_middle_IMA_data[inOffset]);
      break;
    case 6:
      return pgm_read_byte(&whistle_long2_IMA_data[inOffset]);
      break;
    default:
      return 128;//無音
      break;
  }
}

//------------------------------------------------------------------------------------
// ChaN さんのIMA ADPCM decoder を流用

/*-------------------------------------------------*/
/* IMA ADPCM Decoder                               */

//uint8_t ima_decode(    /* Returns the sample value (-32768 to 32767) */
uint16_t ima_decode(    /* Returns the sample value (-32768 to 32767) / Sound extension by MECY 2017/01/14 */
  IMADEC* ad,     /* Work area for the ADPCM stream */
  uint8_t dat,      /* ADPCM data, b3:sign (0:+, 1:-), b2-b0:magnify (0..7) */
  uint8_t hl      /* High or low data read */
)
{
  /*
    char buffer[20];
    sprintf( buffer,"%d,",sample);
    Serial.print(buffer);
  */
  static const uint16_t step[89] __attribute__((__progmem__)) = {
    //  static const uint16_t step[89] = {
    7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 19, 21, 23, 25, 28, 31, 34, 37, 41, 45, 50, 55, 60, 66,
    73, 80, 88, 97, 107, 118, 130, 143, 157, 173, 190, 209, 230, 253, 279, 307, 337, 371, 408, 449,
    494, 544, 598, 658, 724, 796, 876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066, 2272,
    2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358, 5894, 6484, 7132, 7845, 8630, 9493,
    10442, 11487, 12635, 13899, 15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
  };

  int32_t dif;
  uint16_t stp;
  uint8_t idx;

  if (hl == HIGH)   // hl のフラグで、HIGHだったら、上位4bitを下位に移動
    dat = dat >> 4;
  dat = dat & 0x0f;
  /*
    sprintf( buffer,"%d,",dat);
    Serial.print(buffer);
  */
  /* Get current step size */
  idx = ad->idx;            // scale
  stp = pgm_read_word(&step[idx]);
  //  stp = step[idx];

  dif = stp >> 3;   // dif = stp / 8 と等価
  if (dat & 4) dif += stp;
  stp >>= 1;
  if (dat & 2) dif += stp;
  stp >>= 1;
  if (dat & 1) dif += stp;
  if (dat & 8) dif = 1 - dif;
  /* Add the difference to the previous sample */
  dif += ad->val;

  if (dif >  32767) dif =  32767;
  if (dif < -32768) dif = -32768;
  ad->val = (int16_t)dif;

  /* Update step size by magnify */
  dat = (dat & 7) - 4;

  if (dat > 3) {  /* Decrease step size by 1 if magnify is 0..3 */
    if (idx) idx--;
  } else {    /* Increase step size by 2..8 if magnify is 4..7 */
    idx += (dat + 1) << 1;
    if (idx > 88) idx = 88;
  }
  ad->idx = idx;

  //sprintf( buffer,"%#4x:%d",dif,dif);  要素が２つあると、ちゃんと変換できない
  //Serial.println(buffer);
  /*
    sprintf( buffer,"%d,",ad->val);
    Serial.print(buffer);
    sprintf( buffer,"%d,",ad->val/256+128);
    Serial.print(buffer);
    sprintf( buffer,"%d,",idx);
    Serial.println(buffer);
  */
  return ad->val / 256 + 128; /* PCM化した16bitのデータを8bitに間引いて、128のオフセットを履かせています*/
}
//------------------------------------------------------------------------------------



